#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Nov 21 15:21:44 2019

@author: davidblair
"""
import torch
import pyro
import pyro.distributions as dist
from vlpi.utils.UtilityFunctions import random_catcov


class ClinicalDataSimulator:


    def _generate_orthogonal_latent_pheno_params(self):
        effect_precisions=dist.Gamma(self.latentPhenotypeEffectsPrior[0],self.latentPhenotypeEffectsPrior[1]).expand([self.numLatentDimensions]).sample()

        component_assignments = torch.randint(low=0,high=self.numLatentDimensions,size=(self.numObsPhenotypes,))
        pheno_effects=torch.zeros((self.numLatentDimensions,self.numObsPhenotypes),dtype=torch.float32)

        for i in range(self.numLatentDimensions):
            if self.useMonotonic:
                pheno_effects[i,component_assignments==i]=dist.Exponential(torch.sqrt(effect_precisions[i])).expand([1,(component_assignments==i).sum()]).sample()
            else:
                pheno_effects[i,component_assignments==i]=dist.Normal(0.0,1.0/torch.sqrt(effect_precisions[i])).expand([1,(component_assignments==i).sum()]).sample()
        return effect_precisions,pheno_effects


    def _generate_dense_latent_pheno_params(self):
        effect_precisions=dist.Gamma(self.latentPhenotypeEffectsPrior[0],self.latentPhenotypeEffectsPrior[1]).expand([self.numLatentDimensions]).sample()
        pheno_effects=torch.zeros((self.numLatentDimensions,self.numObsPhenotypes),dtype=torch.float32)
        for i in range(self.numLatentDimensions):
            if self.useMonotonic:
                pheno_effects[i]=dist.Exponential(torch.sqrt(effect_precisions[i])).expand([1,self.numObsPhenotypes]).sample()
            else:
                pheno_effects[i]=dist.Normal(0.0,1.0/torch.sqrt(effect_precisions[i])).expand([1,self.numObsPhenotypes]).sample()

        return effect_precisions,pheno_effects


    def _generate_sparse_latent_pheno_params(self,sparsity_rate):
        effect_precisions,pheno_effects = self._generate_dense_latent_pheno_params()
        mask = dist.Bernoulli(probs = torch.ones(pheno_effects.shape)*(1.0-sparsity_rate)).sample().to(torch.float)
        return effect_precisions,pheno_effects*mask



    def __init__(self,numObsPhenotypes,numLatentDimensions,targetDisPrev,isOutlier=False,includeModifierEffects=True,numCatList=[],useMonotonic=True,**kwargs):
        """
        Simulator Class for the variational latent phenotype model.

        Simulated data is based on the simple model outlined in Blair et al.

        Parameters
        ----------
        numObsPhenotypes : int
            Number of observed symptoms/phenotypes to simulate.
        numLatentDimensions : int
            Number of latent/cryptic phenotypes to include into the model.
        targetDisPrev : float
            Prevalence of target/Mendelian disease.
        isOutlier : bool, optional
            Determines whether the target disease is a phenotypic outlier. The default is False.
        includeModifierEffects : bool, optional
            Determines whether or not to include severity-based modifier effects. The default is True.
        numCatList : [int], optional
            List of intergers providing number of categories for each covariate. Total number of covariates: len(numCatList). The default is [].
        useMonotonic : boold, optional
            Whether to enforce monotonicity between latent phenotypes and risk for observed symptoms. The default is True.
        **kwargs : Multiple
            Extensive parameter modification choices. See source code for details. Baseline choices match those used in simulation analyses.

        Returns
        -------
        None.

        """
        allKeywordArgs = list(kwargs.keys())
        self.numObsPhenotypes=numObsPhenotypes
        self.numLatentDimensions=numLatentDimensions
        self.numCatList=numCatList
        self.useMonotonic=useMonotonic
        self.targetDisPrev=torch.tensor(targetDisPrev)
        self.targetDisComponent=torch.randperm(self.numLatentDimensions)[0]
        self.isOutlier=isOutlier
        self.includeModifierEffects=includeModifierEffects

        if 'latentPhenotypeEffectsPrior' not in allKeywordArgs:
            self.latentPhenotypeEffectsPrior=torch.tensor([1.0,1.0],dtype=torch.float32)
        else:
            self.latentPhenotypeEffectsPrior=torch.tensor(kwargs['latentPhenotypeEffectsPrior'],dtype=torch.float32)

        if 'catCovEffectPriors' not in allKeywordArgs:
            self.catCovEffectPriors=torch.tensor([0.0,0.5],dtype=torch.float32)
        else:
            self.catCovEffectPriors=torch.tensor(kwargs['catCovEffectPriors'],dtype=torch.float32)


        if 'interceptPriors' not in allKeywordArgs:
            self.interceptPriors=torch.tensor([-3.0,2.0],dtype=torch.float32)
        else:
            self.interceptPriors=torch.tensor(kwargs['interceptPriors'],dtype=torch.float32)


        if 'targetDisEffectParam' not in allKeywordArgs:
            self.targetDisEffectParam=dist.Normal(0.0,1.0).icdf(1.0-self.targetDisPrev)
        else:
            self.targetDisEffectParam=torch.tensor(kwargs['targetDisEffectParam'],dtype=torch.float32)


        if 'latentPhenotypePriors' not in allKeywordArgs:
            self.latentPhenotypePrior=torch.tensor([0.0,1.0],dtype=torch.float32)
            self.latentPhenotypeOutlierPrior=torch.tensor([0.0,0.0001],dtype=torch.float32)
        else:
            self.latentPhenotypePrior=torch.tensor(kwargs['latentPhenotypePriors'][0],dtype=torch.float32)
            self.latentPhenotypeOutlierPrior=torch.tensor(kwargs['latentPhenotypePriors'][1],dtype=torch.float32)
            print("Warning: By specifying custom latent phenotype priors, you can break the relationship between outliers and spectrums.")

        if self.includeModifierEffects:
            if 'modifierEffectPrior' not in allKeywordArgs:
                self.modifierEffectPrior=torch.tensor([0.0,1.0],dtype=torch.float32)
            else:
                self.modifierEffectPrior=torch.tensor(kwargs['modifierEffectPrior'],dtype=torch.float32)

            if 'modifierEffectThreshold' not in allKeywordArgs:
                self.modifierEffectThreshold=self.targetDisEffectParam
            else:
                self.modifierEffectThreshold=torch.tensor(kwargs['modifierEffectThreshold'],dtype=torch.float32)

        #first construct the latent phenotype effects matrix
        if 'sparsityRate' not in allKeywordArgs:
            sparsityRate = 0.0
        else:
            sparsityRate=kwargs['sparsityRate']
            assert sparsityRate>0.0 and sparsityRate <1.0,"Sparsity rate must lie within (0.0,1.0)"

        if 'orthogonalLatentPhenotypes' not in allKeywordArgs:
            orthogonalLatentPhenotypes=False
        else:
            orthogonalLatentPhenotypes=kwargs['orthogonalLatentPhenotypes']




        if 'latentPhenotypeEffects' not in allKeywordArgs:
            if sparsityRate > 0.0:
                assert orthogonalLatentPhenotypes==False,"Cannot simultaneously use sparse and orthogonal latent phenotypes"
                self.latentPhenotypeEffectsPrecision, self.latentPhenotypeEffects=self._generate_sparse_latent_pheno_params(sparsityRate)
            elif orthogonalLatentPhenotypes==True:
                self.latentPhenotypeEffectsPrecision, self.latentPhenotypeEffects=self._generate_orthogonal_latent_pheno_params()

            else:
                    self.latentPhenotypeEffectsPrecision, self.latentPhenotypeEffects=self._generate_dense_latent_pheno_params()
        else:
            self.latentPhenotypeEffects=torch.tensor(kwargs['latentPhenotypeEffects'])
            assert self.latentPhenotypeEffects.shape[0]==self.numLatentDimensions, "Provided latentPhenotypeEffects matrix does not match numLatentDimensions"
            assert self.latentPhenotypeEffects.shape[1]==self.numObsPhenotypes, "Provided latentPhenotypeEffects matrix does not match numObsPhenotypes"


        #now the intercepts for the observed symptoms
        self.intercepts = dist.Normal(self.interceptPriors[0],self.interceptPriors[1]).expand([self.numObsPhenotypes]).sample()

        #covariate effects (optional)
        if sum(self.numCatList)>0:
            self.covEffects = pyro.sample("covEffects",dist.Normal(self.catCovEffectPriors[0],self.catCovEffectPriors[1]).expand([sum(numCatList),self.numObsPhenotypes]))

    def GenerateClinicalData(self,numSamples):
        """

        Simulates clinical data and returns a dictionary of the results.

        Parameters
        ----------
        numSamples : int
            Number of samples to simulate.

        Returns
        -------
        output : dict
            Dictionary containing data from simulation.

            output['latent_phenotypes']: Siumulated Latent Phenotypes
            output['incidence_data']: Binary symptom matrix
            output['covariate_data']: List of arrays containing categorical covariates.
            output['target_dis_dx']: Mendelian/target disease diagnoses
            output['model_params']: Dictionary of model parameters
            output['model_params']['intercepts']: Vector of intercepts for symptom risk function.
            output['model_params']['latentPhenotypeEffects']: Matrix of symptom risk parameters (loading matrix)

        """

        latentPhenotypes=torch.zeros([numSamples,self.numLatentDimensions],dtype=torch.float32)

        for i in range(self.numLatentDimensions):
            if i!=self.targetDisComponent:
                latentPhenotypes[:,i]=dist.Normal(self.latentPhenotypePrior[0],self.latentPhenotypePrior[1]).expand([numSamples]).sample()
            else:
                targetDisSubjects=dist.Bernoulli(self.targetDisPrev).expand([numSamples]).sample()
                if self.isOutlier==False:
                    latentPhenotypes[:,i]=dist.Normal(self.latentPhenotypePrior[0],self.latentPhenotypePrior[1]).expand([numSamples]).sample()+targetDisSubjects*self.targetDisEffectParam
                else:
                    latentPhenotypes[:,i]=dist.Normal(self.latentPhenotypeOutlierPrior[0],self.latentPhenotypeOutlierPrior[1]).expand([numSamples]).sample()+targetDisSubjects*self.targetDisEffectParam
                if self.includeModifierEffects:
                    latentPhenotypes[:,i]=latentPhenotypes[:,i]+dist.Normal(self.modifierEffectThreshold,1.0).cdf(latentPhenotypes[:,i])*dist.Normal(self.modifierEffectPrior[0],self.modifierEffectPrior[1]).expand([numSamples]).sample()


        cat_cov_list = []
        if sum(self.numCatList)>0:
            for n_cat in self.numCatList:
                cat_cov_list+=[random_catcov(n_cat,numSamples,device='cpu')]
        mean_dx_rates = torch.mm(latentPhenotypes,self.latentPhenotypeEffects)+self.intercepts
        obs_data = dist.Bernoulli(torch.sigmoid(mean_dx_rates)).sample()
        output={}
        output['latent_phenotypes'] = latentPhenotypes
        output['incidence_data'] = obs_data
        output['covariate_data'] = cat_cov_list
        output['target_dis_dx'] = targetDisSubjects
        output['model_params'] = {}
        output['model_params']['intercepts']=self.intercepts
        output['model_params']['latentPhenotypeEffects']=self.latentPhenotypeEffects

        if sum(self.numCatList)>0:
             output['model_params']['covEffects']=self.covEffects

        return output


if __name__=='__main__':

    simulator = ClinicalDataSimulator(20,2,0.001,isOutlier=False)
    simData=simulator.GenerateClinicalData(100000)
    # simLabels = simulator.GenerateTargetDx(simData['latent_phenotypes'])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Aug  9 14:51:17 2019

@author: davidblair
"""

import torch
import torch.nn as nn
import math

class LinearPositiveWeights(nn.Module):

    __constants__ = ['bias', 'in_features', 'out_features']

    def __init__(self, in_features, out_features, bias=True):
        """
        Modifies the Linear neural network class such that the weights are always positive,
        ensuring outputs are a monotonic function of inputs.
        
        """
        super(LinearPositiveWeights, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.log_scale_weight = nn.Parameter(torch.Tensor(out_features, in_features))
        if bias:
            self.bias = nn.Parameter(torch.Tensor(out_features))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        """
        Uses exponential distribution to initialize network weights.
        """
        
        with torch.no_grad():
            self.log_scale_weight.exponential_(100.0)
            self.log_scale_weight=self.log_scale_weight.log_()
            
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(torch.nn.functional.softplus(self.log_scale_weight))
            bound = 1.0 / math.sqrt(fan_in)
            nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, input):
        return nn.functional.linear(input, torch.nn.functional.softplus(self.log_scale_weight), self.bias)

    def extra_repr(self):
        return 'in_features={}, out_features={}, bias={}'.format(
            self.in_features, self.out_features, self.bias is not None
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 22 13:18:43 2019

@author: davidblair
"""
import torch
import pyro
import numpy as np
from vlpi.optim.Optimizer import Optimizer
from vlpi.data.ClinicalDataset import ClinicalDatasetSampler,ClinicalDataset
from vlpi.model.VAE import VAE

__version__ = "0.2.2"

class vLPI:



    def __init__(self,datasetSampler,nLatentDim,latentPhenotypeMap='Linear_Monotonic',**kwargs):
        """
        vLPI (variational-Latent Phenotype Inference) is a statistical model that maps a latent phenotypic space to some vector of observed (binary) traits through a function f(Z), where Z is the latent phenotypic space of interest. Z is assumed to be an isotropic multivariate gaussian distribution, which is equivalent to assuming independence among the latent components. Inference of the model is conducted using a variational approximation of the model marginal model likelihood, which is optimized through stochastic gradient descent. To allow inference to scale to millions of patients and enable subsequent portability, latent phenotype inference is amortized using a non-linear neural network.

        Parameters
        ----------
        datasetSampler : ClinicalDatasetSampler
            Class that contains the ClinicalData and a mechanism for (efficiently) generating random samples.
        nLatentDim : int
            Number of latent phenotypes to include into the model.
        latentPhenotypeMap : str, optional
            Function f(Z) mapping latent to observed phenotypes. Must be one of the following ['Linear','Linear_Monotonic','Nonlinear','Nonlinear_Monotonic']. The default is 'Linear_Monotonic'. Initial testing has only been performed using this function.
        **kwargs: These are arguments passed to the class VAE, which is the modeling class (writtent in PyTorch and pyro) used for inference.

            linkFunction: Function used to map risk-scores produced by f(Z) to the probability scale [0.0,1.0]. Must be in ['Logit','Probit']. Note, 'Logit' is considerably faster.

            dropLinearCovariateColumn: boolean value indicated whether to drop one category from each covariate included into the model. Aids in identifiability. Defaults to True.

            encoderNetworkHyperparameters: dictionary that specifies hyperparameters of the encoder neural networks. Default is a 2-layer MLP with 64 hidden nodes per layer. Dropout rate 0.0 with Batch-Normalization enabled. Larger networks are more difficult to train, while smaller networks are less expressive and may result in poorer approximations of the variational posterior. Based on testing, allowing for dropout appears to limit the models ability to detect low prevalence components. Default: {'n_layers' : 2, 'n_hidden' : 64, 'dropout_rate': 0.0, 'use_batch_norm':True}

            coupleCovariates: Specifies whether to couple covariates to non-linear MLP network (True), or to model them  using an independent linear network (False). Defaults to True. Only relevant if using a Non-linear decoder ('Nonlinear' or 'Nonlinear_Monotonic')

            decoderNetworkHyperparameters: If using a non-linear symptom risk function (f(Z)), this dictionary specifies hyperparameters. Default matches the encoder hyperparameters: {'n_layers' : 2, 'n_hidden' : 64, 'dropout_rate': 0.0, 'use_batch_norm':True}



        Returns
        -------
        None.

        """

        self.sampler = datasetSampler
        if self.sampler.isConditioned:
            self.sampler.ConvertToUnconditional()
            print('Warning: vLPI passed a conditional ClinicalDatasetSampler. Converting to unconditioned sampler.')

        self.all_model_kwargs = kwargs

        n_cat_list = [len(self.sampler.currentClinicalDataset.catCovConversionDicts[x]) for x in self.sampler.includedCovariates]

        assert latentPhenotypeMap in ['Linear','Linear_Monotonic','Nonlinear','Nonlinear_Monotonic'], "Currently supported latent-to-observed phenotype maps include: 'Linear','Linear_Monotonic','Nonlinear','Nonlinear_Monotonic'"

        if latentPhenotypeMap!='Linear_Monotonic':
            print('WARNING: Software is currently optimized for the Linear_Monotonic mapping function. Other mapping functions are experimental, as model optimization parameters were specifically chosen with this mapping function in mind. Properly optimizing these other functions may require experimenting with optimization hyperparameters.')

        self.model = VAE(self.sampler.currentClinicalDataset.numDxCodes, n_cat_list,nLatentDim,latentPhenotypeMap,**kwargs)




    def FitModel(self,batch_size=1000,verbose=False,**kwargs):
        """


        Parameters
        ----------
        batch_size : int, optional
            Size of dataset batches for inference. The default is 1000 patients. Specifying 0 utilizes the full dataset for each optimization step, which is not typically advised due to the approximate nature of the gradient (better to have more frequent updates).
        verbose : bool, optional
            Indicates whether or not to print (to std out) the loss function values and error after every epoch. The default is True.


        Keyword Parameters
        ----------
        maxLearningRate: float, optional
            Specifies the maximum learning rate used during inference. Default is 0.04

        errorTol: float, optional
            Error tolerance in ELBO (computed on held out validation data) to determine convergence. Default is 1e-4.

        numParticles: int, optional
            Number of particles (ie random samples) used to approximate gradient. Default is 1. Computational cost increases linearly with value.

        maxEpochs: int, optional
            Maximum number of epochs (passes through training data) for inference. Note, because annealing and learning rate updates depend on maxEpochs, this offers a simple way to adjust the speed at which these values change.

        computeDevice: int or None, optional
            Specifies compute device for inference. Default is None, which instructs algorithm to use cpu. If integer is provided, then algorithm will be assigned to that integer valued gpu.

        numDataLoaders: int
            Specifies the number of threads used to process data and prepare for upload into the gpu. Note, due to the speed of gpu, inference can become limited by data transfer speed, hence the use of multiple DataLoaders to improve this bottleneck. Default is 0, meaning just the dedicated cpu performs data transfer.

        OneCycleParams: dict with keys 'pctCycleIncrease','initLRDivisionFactor','finalLRDivisionFactor'
            Parameters specifying the One-Cycle learning rate adjustment strategy, which helps to enable good anytime performance.
            pctCycleIncrease--fraction of inference epochs used for increasing learning rate. Default: 0.1
            initLRDivisionFactor--initial learning rate acheived by dividing maxLearningRate by this value. Default: 25.0
            finalLRDivisionFactor--final learning rate acheived by dividing maxLearningRate by this value. Default: 1e4


        KLAnnealingParams: dict with keys 'initialTemp','maxTemp','fractionalDuration','schedule'
            Parameters that define KL-Annealing strategy used during inference, important for avoiding local optima. Note, annealing is only used for computation of ELBO and gradients on training data. Validation data ELBO evaluation, used to monitor convergence, is performed at the maximum desired temperature (typically 1.0, equivalent to standard variational inference). Therefore, it is possible for the model to converge even when the temperature hasn't reached it's final value. It's possible that further cooling would find a better optimum, but this is highly unlikely in practice.
            initialTemp--initial temperature during inference. Default: 0.0
            maxTemp--final temperature obtained during inference. Default: 1.0 (standard variational inference)
            fractionalDuration--fraction of inference epochs used for annealing. Default is 0.25
            schedule--function used to change temperature during inference. Defualt is 'cosine'. Options: 'cosine','linear'

        'OpimizationStrategy': str, optional
            Specifies a strategry for optimization. Options include: 'Full','DecoderOnly','EncoderOnly'. Useful for debugging. Default is 'Full'


        Returns
        -------
        output : list
            List containing the following information: [loss function value of best model (computed on validation data),sequence of training loss values, sequence of validation loss values, error estimates across iterations (computed on validation data)].



        """


        ######### Parse Keyword Arguments #########
        allKeywordArgs = list(kwargs.keys())


        if 'maxLearningRate' in allKeywordArgs:
            maxLearningRate=kwargs['maxLearningRate']
        else:
            maxLearningRate=0.04


        if 'errorTol' in allKeywordArgs:
            errorTol=kwargs['errorTol']
        else:
            errorTol=1e-4

        if 'numParticles' in allKeywordArgs:
            numParticles=kwargs['numParticles']
        else:
            numParticles=1


        if 'maxEpochs' in allKeywordArgs:
            maxEpochs=kwargs['maxEpochs']
        else:
            maxEpochs=200


        if 'computeDevice' in allKeywordArgs:
            computeDevice=kwargs['computeDevice']
        else:
            computeDevice=None

        if 'numDataLoaders' in allKeywordArgs:
            numDataLoaders=kwargs['numDataLoaders']
            if computeDevice in [None,'cpu']:
                assert numDataLoaders==0,"Specifying number of dataloaders other than 0 only relevant when using GPU computing"
        else:
            numDataLoaders=0

        if 'OneCycleParams' in allKeywordArgs:
            OneCycleParams=kwargs['OneCycleParams']
            assert set(OneCycleParams.keys())==set(['pctCycleIncrease','initLRDivisionFactor','finalLRDivisionFactor']),"One-cycle LR scheduler requires the following parameters:'pctCycleIncrease','initLRDivisionFactor','finalLRDivisionFactor'"

        else:
            OneCycleParams={'pctCycleIncrease':0.1,'initLRDivisionFactor':25.0,'finalLRDivisionFactor':1e4}

        if 'KLAnnealingParams' in allKeywordArgs:
            KLAnnealingParams=kwargs['KLAnnealingParams']
            assert set(KLAnnealingParams.keys())==set(['initialTemp','maxTemp','fractionalDuration','schedule']),"KL Annealing Parameters must be dictionary with the following keys: 'initialTemp','maxTemp','fractionalDuration','schedule'"
        else:
            KLAnnealingParams={'initialTemp':0.0,'maxTemp':1.0,'fractionalDuration':0.25,'schedule': 'cosine'}

        if 'OpimizationStrategy' in allKeywordArgs:
            optimStrategy = kwargs['OpimizationStrategy']
            assert optimStrategy in ['Full','DecoderOnly','EncoderOnly'],"Available optimization strategies include: Full, DecoderOnly, EncoderOnly"
        else:
            optimStrategy='Full'


        pyro.clear_param_store()

        optimizer=Optimizer(self.model,self.sampler,optimizationParameters={'maxLearningRate': maxLearningRate,'maxEpochs': maxEpochs,'numParticles':numParticles},computeConfiguration={'device':computeDevice,'numDataLoaders':numDataLoaders},OneCycleParams=OneCycleParams,KLAnnealingParams=KLAnnealingParams)
        if batch_size > 0:
            output=optimizer.BatchTrain(batch_size,errorTol=errorTol,optimizationStrategy=optimStrategy,verbose=verbose)
        else:
            output=optimizer.FullDatasetTrain(errorTol=errorTol,optimizationStrategy=optimStrategy,verbose=verbose)
        print('Inference complete. Final Loss: {0:10.2f}'.format(output[0]))
        return output



    def ComputeEmbeddings(self,dataArrays=None,randomize=False,returnStdErrors=False):
        """
        Returns low dimnsional embeddings for the dataset. By default, perfoms
        operation on full testing and training dataset contained withinin sampler
        unless dataArrays are provided, in which case embeddings are computed
        for the provided arrays.


        Parameters
        ----------
        dataArray : 2-tuple of array-like data structures, corresponding to
        (incicidence array, [list of categorical covariates]), optional
            Can provide custom arrays not contained within ClinicalDatasetSampler
            in order to compute embeddings. The default value is None.
        randomize : bool, optional
            Determines whether to shuffle training/testing data prior to
            computation over data contained within ClinicalDatasetSampler.
            The default is False.
        returnStdErrors: bool, optional
            Indicates whether to return std errors for the embeddings in addition to the means.

        Returns
        -------

        Note: If std errors are included, then each of the below returned arrays will have 3-dimensions, wihth the third providing [mean,scale].

        numpy array
            training set embeddings.
        numpy array
            valdiation set embeddings.

        -- or --

        numpy array
            embeddings for dataArrays

        """

        if dataArrays is None:

            previousArrayType = self.sampler.returnArrays
            if self.sampler.returnArrays!='Torch':
                self.sampler.ChangeArrayType('Torch')

            trainingData=self.sampler.ReturnFullTrainingDataset(randomize=randomize)
            testingData=self.sampler.ReturnFullTestingDataset(randomize=randomize)

            train_embed = self.model.PredictLatentPhenotypes(trainingData[0],trainingData[1],returnScale=returnStdErrors)
            test_embed = self.model.PredictLatentPhenotypes(testingData[0],testingData[1],returnScale=returnStdErrors)

            if previousArrayType!='Torch':
                self.sampler.ChangeArrayType(previousArrayType)

            if returnStdErrors:

                return np.concatenate((train_embed[0].detach().numpy()[:,:,np.newaxis],train_embed[1].detach().numpy()[:,:,np.newaxis]), axis=-1),np.concatenate((test_embed[0].detach().numpy()[:,:,np.newaxis],test_embed[1].detach().numpy()[:,:,np.newaxis]),axis=-1)
            else:
                return train_embed.detach().numpy(),test_embed.detach().numpy()

        else:
            incidence_array = dataArrays[0]

            if torch.is_tensor(incidence_array) is False:
                incidence_array=self.sampler._torchWrapper(incidence_array)

            list_cov_arrays=[]
            for x in dataArrays[1]:
                if torch.is_tensor(x) is False:
                    x=self.sampler._torchWrapper(x)
                list_cov_arrays+=[x]

            embeddings=self.model.PredictLatentPhenotypes(incidence_array, list_cov_arrays,returnScale=returnStdErrors)

            if returnStdErrors:
                return np.concatenate((embeddings[0].detach().numpy()[:,:,np.newaxis],embeddings[1].detach().numpy()[:,:,np.newaxis]),axis=-1)
            else:
                return embeddings.detach().numpy()



    def ReturnComponents(self):
        """
        Returns the matrix of parameters defining f(Z), the function mapping the latent phenotypic space to the observed symptom risk values (sometimes called 'loadings' in factor analysis).
        This function is only available for linear f(Z).

        Returns
        -------
        numpy.array
            Matrix of latent phenotype loadings.

        """

        assert self.model.decoderType in ['Linear_Monotonic','Linear'],"Components only availabe for Linear models. vLPI fit using a non-linear mapping function."
        mapping_func_state_dict = self.model.decoder.state_dict()

        if self.model.decoderType == 'Linear_Monotonic':
            return torch.nn.functional.softplus(mapping_func_state_dict['linear_latent.log_scale_weight']).detach().numpy().T
        else:
            return mapping_func_state_dict['linear_latent.weight'].detach().numpy().T



    def ComputePerplexity(self,dataArrays=None, randomize=False):
        """
        Computes the per-datum perplexity (-1.0*ELBO) of the training/test data or the provided dataArrays.
        Can be thought of as a type of reconstruction error (sometimes also referred to as data "surprisal")

        Parameters
        ----------
        dataArray : 2-tuple of array-like data structures, corresponding to
        (incicidence array, [list of categorical covariates]), optional
            Can provide custom arrays not contained within ClinicalDatasetSampler
            in order to compute embeddings. The default value is None.
        randomize : bool, optional
            Determines whether to shuffle training/testing data prior to
            computation over data contained within ClinicalDatasetSampler.
            The default is False.

        Returns
        -------
        numpy.array
            Array of per-datum perplexity values.

        """
        if dataArrays is None:
            previousArrayType = self.sampler.returnArrays
            if self.sampler.returnArrays!='Torch':
                self.sampler.ChangeArrayType('Torch')

            trainingData=self.sampler.ReturnFullTrainingDataset(randomize=randomize)
            testingData=self.sampler.ReturnFullTestingDataset(randomize=randomize)

            train_elbo = self.model.ComputeELBOPerDatum(trainingData[0],trainingData[1])
            test_elbo = self.model.ComputeELBOPerDatum(testingData[0],testingData[1])

            if previousArrayType!='Torch':
                self.sampler.ChangeArrayType(previousArrayType)
            return -1.0*train_elbo.detach().numpy(),-1.0*test_elbo.detach().numpy()
        else:
            incidence_array = dataArrays[0]

            if torch.is_tensor(incidence_array) is False:
                incidence_array=self.sampler._torchWrapper(incidence_array)

            list_cov_arrays=[]
            for x in dataArrays[1]:
                if torch.is_tensor(x) is False:
                    x=self.sampler._torchWrapper(x)
                list_cov_arrays+=[x]
            elbo=self.model.ComputeELBOPerDatum(incidence_array, list_cov_arrays)
            return -1.0*elbo.detach().numpy()


    def _readModelFromFile(self,fName):
        with open(fName,'rb') as f:
            model_dict = torch.load(f,map_location='cpu')
        return model_dict


    def LoadModel(self,stored_model):
        """
        Loads previously fit model either from a dictionary (generated using PackageModel) or from a file path (with file constructed using PackageModel)

        Parameters
        ----------
        stored_model : either dict or str (file path)

        Returns
        -------
        None.

        """
        if not isinstance(stored_model,dict):
            assert isinstance(stored_model,str),"Expects file name if not provided with dictionary."
            stored_model = self._readModelFromFile(stored_model)
        assert set(stored_model.keys())==set(['model_state','meta_data']),"Model dictionary must contain the following elements: 'model_state','meta_data'"

        #first load meta data, make sure it aligns with structure of current vlpi instance

        self.model = VAE(stored_model['meta_data']['numObsTraits'], stored_model['meta_data']['numCatList'],stored_model['meta_data']['nLatentDim'],stored_model['meta_data']['latentPhenotypeMap'],**self.all_model_kwargs)
        self.model.LoadPriorState(stored_model)



    def PackageModel(self,fName=None):
        """
        Packages the current model and returns it as a python dictionary. Will optionally write this dictionary to disk using PyTorch.

        Parameters
        ----------
        fName : str, default None
            File path to save model to disk. The default is None, which means that only a model dictionary will be returned.

        Returns
        -------
        model_dict : dict
            Dictionary containing fitted model parameters in addition to general meta data.

        """
        model_dict = self.model.PackageCurrentState()
        model_dict['meta_data']={}
        model_dict['meta_data']['numObsTraits']=self.model.numObsTraits
        model_dict['meta_data']['numCatList']=self.model.numCatList
        model_dict['meta_data']['latentPhenotypeMap']=self.model.decoderType
        model_dict['meta_data']['nLatentDim']=self.model.nLatentDim

        if fName is not None:
            with open(fName,'wb') as f:
                torch.save(model_dict,f)
        return model_dict



from vlpi.data.ClinicalDataset import ClinicalDataset,ClinicalDatasetSampler
from vlpi.data.ClinicalDataSimulator import ClinicalDataSimulator
from vlpi.vLPI import vLPI
import torch
import string
import numpy as np

seed=torch.manual_seed(1023)

numberOfSamples=50000
numberOfSymptoms=20

rareDiseaseFrequency=0.001
numLatentPhenotypes=2
simulator = ClinicalDataSimulator(numberOfSymptoms,numLatentPhenotypes,rareDiseaseFrequency,numCatList=[2,3,4])
simulatedData=simulator.GenerateClinicalData(numberOfSamples)
symptomConversionMap=dict(zip(string.ascii_uppercase[0:numberOfSymptoms],range(0,numberOfSymptoms)))
symptomConversionMapInv=dict(zip(range(0,numberOfSymptoms),string.ascii_uppercase[0:numberOfSymptoms]))
simulatedIndex = list([str(i) for i in range(numberOfSamples)])
np.random.shuffle(simulatedIndex)


### write simulated data to file ###
f1=open('SimDxFile_NoCovariates.txt','w')
f2=open('SimDxFile_WithCovariates.txt','w')
f1.write('p_id\tdx\n')
f2.write('p_id\tc1\tc2\tc3\tdx\n')
for i in range(numberOfSamples):
	f1.write('{0:s}\t{1:s}\n'.format(simulatedIndex[i],','.join([symptomConversionMapInv[int(s)] for s in torch.where(simulatedData['incidence_data'][i])[0]])))
	f2.write('{0:s}\t{1:d}\t{2:d}\t{3:d}\t{4:s}\n'.format(simulatedIndex[i],int(simulatedData['covariate_data'][0][i]),int(simulatedData['covariate_data'][1][i]),int(simulatedData['covariate_data'][2][i]),','.join([symptomConversionMapInv[int(s)] for s in torch.where(simulatedData['incidence_data'][i])[0]])))
f1.close()
f2.close()


## load the data into a ClinicalDataset class
clinicalData = ClinicalDataset(symptomConversionMap)
# clinicalData.ReadDatasetFromFile('SimDxFile_NoCovariates.txt',1,indexColumn=0)
clinicalData.ReadDatasetFromFile('SimDxFile_WithCovariates.txt',4,indexColumn=0)
# clinicalData.LoadFromArrays(simulatedData['incidence_data'],subjectIDs=simulatedIndex,covariateArrays=dict(zip(['c1','c2','c3'],simulatedData['covariate_data'])))

# #change the dataset to have only 20 symptoms (defaults to full ICD10-CM codebook), named after the first 20 letters

# # now load into the data structure. Note, when using in this manner, it's the users responsibility to make sure that the input data columns match the data columns of the ClinicalDataset.
# clinicalData.LoadFromArrays(simulatedData['incidence_data'],simulatedData['covariate_data'],[],catCovDicts=None, arrayType = 'Torch')

# ## Now load the ClincalDataset into ClinicalDatasetSampler
training_data_fraction=0.75
sampler = ClinicalDatasetSampler(clinicalData,training_data_fraction,returnArrays='Torch')

# ## Intantation the models
infNumberOfLatentPhenotypes=10
vlpiModel= vLPI(sampler,infNumberOfLatentPhenotypes)

inference_output = vlpiModel.FitModel(batch_size=1000,errorTol=(1.0/numberOfSamples))
# vlpiModel.PackageModel('ExampleModel.pth')

# inferredCrypticPhenotypes=vlpiModel.ComputeEmbeddings((simulatedData['incidence_data'],simulatedData['covariate_data']))
# riskFunction=vlpiModel.ReturnComponents()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Nov 21 11:34:37 2019

@author: davidblair
"""
from sklearn.feature_selection import SelectFdr,chi2,f_regression
from vlpi.utils.UtilityFunctions import fisher_exact,T_test
import numpy as np


class FeatureSelection:


    def __init__(self,datasetSampler):
        """
        Class used to perform feature selection on the clinical dataset, passed through using a ClinicalDatasetSampler.

        Parameters
        ----------
        datasetSampler : vlpi.data.ClinicalDatasetSampler
            ClinicalDatasetSampler class for the dataset being analyzed.

        Returns
        -------
        None
        """


        self.sampler=datasetSampler

    def SelectComorbidTraits_ContinuousFeature(self,featureVector,FDR,modifyDataset=False,use_ttest=False):
        """
        Selects features correlated with some continuous variable.

        Parameters
        ----------
        featureVector : [float]
            Vector of floating values for feature selection. Must be sorted in the same order as the index for the ClinicalDatasetSampler training dataset.
        FDR : float
            False discovery rate cut off for feature selection
        modifyDataset : bool
            If True, then features that faile to be selected will be dropped from the dataset.
        use_ttest : bool
            By default, uses F-test to estimate correlation between featureVector and features. If True, instead uses T-test to perform association.

        Returns
        -------
        tuple of arrays
            (Index of selected features, Feature Scores ,Feature P-values)

        """


        previousArrayType = self.sampler.returnArrays
        if self.sampler.returnArrays!='Sparse':
            self.sampler.ChangeArrayType('Sparse')

        sparseTrainingData=self.sampler.ReturnFullTrainingDataset(randomize=False)
        dataMatrix=sparseTrainingData[0]

        if use_ttest:
            fdr=SelectFdr(T_test, alpha=FDR)
        else:
            fdr=SelectFdr(f_regression, alpha=FDR)

        fdr_fit = fdr.fit(dataMatrix,featureVector.ravel())
        discIndx=np.where(fdr_fit.get_support()==True)[0]


        if modifyDataset:
            self.sampler.currentClinicalDataset.IncludeOnly([self.sampler.currentClinicalDataset.dataIndexToDxCodeMap[x] for x in discIndx])

        if previousArrayType!='Sparse':
            self.sampler.ChangeArrayType(previousArrayType)

        return discIndx, fdr_fit.scores_[discIndx],fdr_fit.pvalues_[discIndx]

    def SelectComorbidTraits(self,FDR,modifyDataset=False,useChi2=True):
        """
        Selects features (symptoms) correlated with some dichotomous variable (disease diagnosis), hence co-morbid. This dichotomous variable is automatically inferred from ClinicalDatasetSampler, as it is whatever the sampler is conditioned on.

        Parameters
        ----------

        FDR : float
            False discovery rate cut off for feature selection
        modifyDataset : bool
            If True, then features that faile to be selected will be dropped from the dataset.
        useChi2 : bool
            By default, uses chi-sq test to estimate co-morbidity between featureVector and features. If False, then Fisher's exact test is used. 

        Returns
        -------
        tuple of arrays
            (Index of selected features, Feature Scores ,Feature P-values)

        """


        assert self.sampler.isConditioned==True,"Cannot perform feature selection without being conditioned on some disease of interest"
        previousArrayType = self.sampler.returnArrays
        if self.sampler.returnArrays!='Sparse':
            self.sampler.ChangeArrayType('Sparse')


        sparseTrainingData=self.sampler.ReturnFullTrainingDataset(randomize=False)
        dataMatrix=sparseTrainingData[0]
        incidenceVec =sparseTrainingData[2]

        if useChi2==False:
            fdr=SelectFdr(fisher_exact, alpha=FDR)
        else:
            fdr=SelectFdr(chi2, alpha=FDR)


        fdr_fit = fdr.fit(dataMatrix,incidenceVec.toarray())
        discIndx=np.where(fdr_fit.get_support()==True)[0]

        if modifyDataset:
            self.sampler.currentClinicalDataset.IncludeOnly([self.sampler.currentClinicalDataset.dataIndexToDxCodeMap[x] for x in discIndx])

        if previousArrayType!='Sparse':
            self.sampler.ChangeArrayType(previousArrayType)

        return discIndx, fdr_fit.scores_[discIndx],fdr_fit.pvalues_[discIndx]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 12 12:31:35 2019

@author: davidblair
"""

import torch
import torch.nn as nn
from typing import Iterable
from vlpi.model.FCLayers import FCLayers
import collections


class MeanScaleEncoder(nn.Module):
    r"""
    Adapted from the scVI module: https://github.com/YosefLab/scVI


    Encodes data of ``n_input`` dimensions into a latent space of ``n_output``
    dimensions using a fully-connected neural network of ``n_hidden`` layers.
    :param n_input: The dimensionality of the input (data space)
    :param n_output: The dimensionality of the output (latent space)
    :param n_cat_list: A list containing the number of categories
                       for each category of interest. Each category will be
                       included using a one-hot encoding
    :param n_layers: The number of fully-connected hidden layers
    :param n_hidden: The number of nodes per hidden layer
    :dropout_rate: Dropout rate to apply to each of the hidden layers
    """

    def __init__(self, n_input: int, n_output: int,
                n_cat_list: Iterable[int] = None, n_layers: int = 2,
                 n_hidden: int = 128, dropout_rate: float = 0.1, use_batch_norm:bool=True):

        super(MeanScaleEncoder,self).__init__()

        self.encoder = FCLayers(n_in=n_input, n_out=n_hidden, n_cat_list=n_cat_list, n_layers=n_layers,
                                n_hidden=n_hidden, dropout_rate=dropout_rate,use_batch_norm=use_batch_norm)
        self.mean_encoder = nn.Linear(n_hidden, n_output,bias=True)
        self.var_encoder = nn.Linear(n_hidden, n_output,bias=True)

    def forward(self, x: torch.Tensor, *cat_list: int):
        r"""The forward computation for a single sample.
         #. Encodes the data into latent space using the encoder network
        :param x: tensor with shape (n_input,)
        :param cat_list: list of category membership(s) for this sample
        """

        # Parameters for latent distribution
        q = self.encoder(x, *cat_list)
        q_m = self.mean_encoder(q)
        q_v = torch.exp(self.var_encoder(q)) + 1e-6
        return q_m, q_v
